
from logging import Logger
from logging import getLogger

from pyorthogonalrouting.Common import Integers
from pyorthogonalrouting.Common import integerListFactory

from pyorthogonalrouting.Functions import computePt
from pyorthogonalrouting.Functions import extrudeConnectorPoint
from pyorthogonalrouting.Functions import isVerticalSide
from pyorthogonalrouting.Functions import makePt

from pyorthogonalrouting.ConnectorPoint import ConnectorPoint
from pyorthogonalrouting.Functions import simplifyPaths
from pyorthogonalrouting.Line import Lines
from pyorthogonalrouting.OrthogonalConnectorByProduct import OrthogonalConnectorByProduct

from pyorthogonalrouting.OrthogonalConnectorOptions import OrthogonalConnectorOptions

from pyorthogonalrouting.Grid import Grid
from pyorthogonalrouting.Point import Point
from pyorthogonalrouting.Point import Points
from pyorthogonalrouting.Point import pointsFactory
from pyorthogonalrouting.PointGraph import GraphAndConnections
from pyorthogonalrouting.PointGraph import PointGraph

from pyorthogonalrouting.Rectangle import Rectangle
from pyorthogonalrouting.Rectangle import Rectangles
from pyorthogonalrouting.enumerations.Side import Side


class OrthogonalConnector:
    """
    https://medium.com/swlh/routing-orthogonal-diagram-connectors-in-javascript-191dc2c5ff70
    https://gist.github.com/jose-mdz/4a8894c152383b9d7a870c24a04447e4

    """

    byProduct: OrthogonalConnectorByProduct = OrthogonalConnectorByProduct()

    def __init__(self):
        self.logger: Logger = getLogger(__name__)

    @classmethod
    def route(cls, options: OrthogonalConnectorOptions) -> Points:

        pointA:             ConnectorPoint = options.pointA
        pointB:             ConnectorPoint = options.pointB
        globalBoundsMargin: int            = options.globalBoundsMargin

        spots:       Points   = pointsFactory()
        verticals:   Integers = integerListFactory()
        horizontals: Integers = integerListFactory()

        sideA: Side = pointA.side
        sideB: Side = pointB.side

        sideAVertical: bool = isVerticalSide(sideA)
        sideBVertical: bool = isVerticalSide(sideB)

        originA: Point = computePt(pointA)
        originB: Point = computePt(pointB)

        shapeA:    Rectangle = Rectangle.fromRect(r=pointA.shape)
        shapeB:    Rectangle = Rectangle.fromRect(r=pointB.shape)
        bigBounds: Rectangle = Rectangle.fromRect(r=options.globalBounds)

        shapeMargin: int = options.shapeMargin

        inflatedA: Rectangle = shapeA.inflate(horizontal=shapeMargin, vertical=shapeMargin)
        inflatedB: Rectangle = shapeB.inflate(horizontal=shapeMargin, vertical=shapeMargin)

        # Check bounding boxes collision

        if inflatedA.intersects(rectangle=inflatedB):
            shapeMargin = 0
            inflatedA = shapeA
            inflatedB = shapeB

        inflatedBounds = inflatedA.union(inflatedB).inflate(globalBoundsMargin, globalBoundsMargin)

        # Curated bounds to stick to
        bounds: Rectangle = Rectangle.fromLTRB(
            left=max(inflatedBounds.left, bigBounds.left),
            top=max(inflatedBounds.top, bigBounds.top),
            right=min(inflatedBounds.right, bigBounds.right),
            bottom=min(inflatedBounds.bottom, bigBounds.bottom)
        )

        # Add edges to rulers
        for b in [inflatedA, inflatedB]:
            verticals.append(b.left)
            verticals.append(b.right)
            horizontals.append(b.top)
            horizontals.append(b.bottom)

        # Rulers at origins of shapes
        # (sideAVertical ? verticals : horizontals).push(sideAVertical ? originA.x : originA.y);
        # (sideBVertical ? verticals : horizontals).push(sideBVertical ? originB.x : originB.y);
        # Typescript is too cute

        if sideAVertical is True:
            verticals.append(originA.x)
        else:
            horizontals.append(originA.y)
        if sideBVertical is True:
            verticals.append(originB.x)
        else:
            horizontals.append(originB.y)

        # const add = (dx: number, dy: number) => spots.push(makePt(p.x + dx, p.y + dy));
        def add(pt: Point, dx: int, dy: int):
            spots.append(makePt(pt.x + dx, pt.y + dy))

        # Points of shape antennas
        for connectorPt in [pointA, pointB]:
            p: Point = computePt(p=connectorPt)
            match connectorPt.side:
                case Side.TOP:
                    add(p, dx=0, dy=-shapeMargin)
                case Side.RIGHT:
                    add(p, dx=shapeMargin, dy=0)
                case Side.BOTTOM:
                    add(p, dx=0, dy=shapeMargin)
                case Side.LEFT:
                    add(p, dx=-shapeMargin, dy=0)
                case _:
                    assert False, f'I do not understand that side {connectorPt.side=}'

        # Sort rulers
        verticals.sort()
        horizontals.sort()

        # Create grid
        grid:       Grid   = Grid.rulersToGrid(verticals=verticals, horizontals=horizontals, bounds=bounds)
        gridPoints: Points = Grid.gridToSpots(grid=grid, obstacles=Rectangles([inflatedA, inflatedB]))

        # Add to spots
        spots = Points(spots + gridPoints)

        # Create graph
        graphAndConnections: GraphAndConnections = PointGraph.createGraph(spots)
        graph:       PointGraph = graphAndConnections.graph
        connections: Lines      = graphAndConnections.connections

        # Origin and destination by extruding antennas
        origin:      Point = extrudeConnectorPoint(pointA, shapeMargin)
        destination: Point = extrudeConnectorPoint(pointB, shapeMargin)

        start: Point = computePt(pointA)
        end:   Point = computePt(pointB)

        OrthogonalConnector.byProduct.spots       = spots
        OrthogonalConnector.byProduct.vRulers     = verticals
        OrthogonalConnector.byProduct.hRulers     = horizontals
        OrthogonalConnector.byProduct.grid        = grid.rectangles
        OrthogonalConnector.byProduct.connections = connections

        path: Points = PointGraph.shortestPath(graph, origin, destination)

        if len(path) > 0:
            pathToSimplify: Points = Points(Points([start]) + path + Points([end]))
            return simplifyPaths(points=pathToSimplify)
        else:
            return Points([])




from typing import List
from typing import NewType

from dataclasses import dataclass
from typing import Tuple
from typing import cast

from pyorthogonalrouting.Common import NOT_SET_INT


@dataclass(eq=True)
class Point:
    """
    A point in space.
    """
    x: int = NOT_SET_INT
    y: int = NOT_SET_INT

    def toTuple(self) -> Tuple[int, int]:
        return self.x, self.y

    def __hash__(self) -> int:
        return hash((self.x, self.y))


Points = NewType('Points', List[Point])


def pointsFactory() -> Points:
    return Points([])


NO_POINT: Point = cast(Point, None)




from typing import List
from typing import NewType

from dataclasses import dataclass
from dataclasses import field

from pyorthogonalrouting.Common import Integers
from pyorthogonalrouting.Common import integerListFactory

from pyorthogonalrouting.Line import Lines
from pyorthogonalrouting.Line import linesFactory

from pyorthogonalrouting.Point import Points
from pyorthogonalrouting.Point import pointsFactory

from pyorthogonalrouting.Rectangle import Rectangles
from pyorthogonalrouting.Rectangle import rectanglesFactory


@dataclass
class OrthogonalConnectorByProduct:

    hRulers:     Integers   = field(default_factory=integerListFactory)
    vRulers:     Integers   = field(default_factory=integerListFactory)
    spots:       Points     = field(default_factory=pointsFactory)
    grid:        Rectangles = field(default_factory=rectanglesFactory)
    connections: Lines      = field(default_factory=linesFactory)


OrthogonalConnectorByProducts = NewType('OrthogonalConnectorByProducts', List[OrthogonalConnectorByProduct])




from dataclasses import dataclass

from pyorthogonalrouting.Rect import Rect
from pyorthogonalrouting.Rect import NO_RECT

from pyorthogonalrouting.ConnectorPoint import ConnectorPoint
from pyorthogonalrouting.ConnectorPoint import NO_CONNECTOR_POINT


@dataclass(slots=True)
class OrthogonalConnectorOptions:

    pointA:             ConnectorPoint = NO_CONNECTOR_POINT
    pointB:             ConnectorPoint = NO_CONNECTOR_POINT
    shapeMargin:        int            = 20
    globalBoundsMargin: int            = 50
    globalBounds:       Rect           = NO_RECT




from typing import List
from typing import NewType
from typing import cast

from dataclasses import dataclass

from pyorthogonalrouting.Common import NOT_SET_INT
from pyorthogonalrouting.Size import Size

DELIMITER: str = ','


@dataclass
class Rect(Size):
    """
    Represents a Rectangle by location and size
    """
    left: int = NOT_SET_INT
    top:  int = NOT_SET_INT

    @classmethod
    def deSerialize(cls, value: str) -> 'Rect':

        values: List[str] = value.split(sep=DELIMITER)

        assert len(values) == 4, 'Incorrectly formatted `Rect` values'

        rect: Rect = Rect()
        rect.top    = int(values[0])
        rect.left   = int(values[1])
        rect.width  = int(values[2])
        rect.height = int(values[3])

        return rect

    def __str__(self):

        sizeStr: str = super().__str__()
        return f'{self.left}{DELIMITER}{self.top}{DELIMITER}{sizeStr}'

    def __repr__(self):
        return self.__str__()


Rects = NewType('Rects', List[Rect])


NO_RECT: Rect = cast(Rect, None)




from logging import Logger
from logging import getLogger

from codeallybasic.DynamicConfiguration import DynamicConfiguration
from codeallybasic.DynamicConfiguration import KeyName
from codeallybasic.DynamicConfiguration import SectionName
from codeallybasic.DynamicConfiguration import Sections
from codeallybasic.DynamicConfiguration import ValueDescription
from codeallybasic.DynamicConfiguration import ValueDescriptions

from codeallybasic.SecureConversions import SecureConversions

from codeallybasic.SingletonV3 import SingletonV3

from pyorthogonalrouting.Rect import Rect

DEFAULT_SHAPE_MARGIN:              str  = '20'
DEFAULT_GLOBAL_BOUNDS_MARGIN:      str  = '50'
DEFAULT_SOURCE_EDGE_DISTANCE:      str  = '0.5'
DEFAULT_DESTINATION_EDGE_DISTANCE: str  = '0.5'
DEFAULT_GLOBAL_BOUNDS:             str  = Rect(left=0, top=0, width=500, height=500).__str__()

#         ConfigurationNameValue(name=PropertyName('destinationEdgeDistance'), defaultValue=DEFAULT_DESTINATION_EDGE_DISTANCE),

sectionMain: ValueDescriptions = ValueDescriptions(
    {
        KeyName('shapeMargin'):             ValueDescription(defaultValue=DEFAULT_SHAPE_MARGIN,              deserializer=SecureConversions.secureInteger),
        KeyName('globalBoundsMargin'):      ValueDescription(defaultValue=DEFAULT_GLOBAL_BOUNDS_MARGIN,      deserializer=SecureConversions.secureInteger),
        KeyName('globalBounds'):            ValueDescription(defaultValue=DEFAULT_GLOBAL_BOUNDS,             deserializer=Rect.deSerialize),
        KeyName('sourceEdgeDistance'):      ValueDescription(defaultValue=DEFAULT_SOURCE_EDGE_DISTANCE,      deserializer=SecureConversions.secureFloat),
        KeyName('destinationEdgeDistance'): ValueDescription(defaultValue=DEFAULT_DESTINATION_EDGE_DISTANCE, deserializer=SecureConversions.secureFloat),
    }
)

PY_ORTHOGONAL_ROUTING_SECTIONS: Sections = Sections(
    {
        SectionName('Main'): sectionMain,
    }
)


class Configuration(DynamicConfiguration, metaclass=SingletonV3):

    def __init__(self):
        """
        Make the logger protected to avoid infinite access loop in Dynamic Configuration
        """
        self._logger: Logger = getLogger(__name__)

        super().__init__(baseFileName='pyorthogonalrouting.ini', moduleName='pyorthogonalrouting', sections=PY_ORTHOGONAL_ROUTING_SECTIONS)




from typing import List
from typing import NewType

from dataclasses import dataclass

from pyorthogonalrouting.Point import NO_POINT
from pyorthogonalrouting.Point import Point


@dataclass
class Line:
    """
    A line between two points.
    """
    a: Point = NO_POINT
    b: Point = NO_POINT


Lines = NewType('Lines', List[Line])


def linesFactory() -> Lines:
    return Lines([])




from math import sqrt

from pyorthogonalrouting.ConnectorPoint import ConnectorPoint

from pyorthogonalrouting.Point import Point
from pyorthogonalrouting.Point import Points
from pyorthogonalrouting.PointNode import Distance

from pyorthogonalrouting.Rectangle import Rectangle
from pyorthogonalrouting.enumerations.BendDirection import BendDirection

from pyorthogonalrouting.enumerations.Side import Side


def distance(a: Point, b: Point) -> Distance:

    intDistance: int = round(sqrt(pow(b.x - a.x, 2) + pow(b.y - a.y, 2)))

    return Distance(intDistance)


def isVerticalSide(side: Side) -> bool:
    """

    Args:
        side:

    Returns: 'True' if the side belongs on the vertical axis, else it returns 'False'
    """
    return side == Side.TOP or side == Side.BOTTOM


def makePt(x: int, y: int) -> Point:
    """
    Utility Point creator

    Args:
        x:
        y:

    Returns:

    """
    return Point(x=x, y=y)


def computePt(p: ConnectorPoint) -> Point:
    """
    TODO: rename to toPoint
    Args:
        p:  Gets the actual point of the connector based on the distance parameter

    Returns:

    """
    b: Rectangle = Rectangle.fromRect(p.shape)

    match p.side:
        case Side.BOTTOM:
            return makePt(b.left + round (b.width * p.distance), b.bottom)
        case Side.TOP:
            return makePt(b.left + round(b.width * p.distance), b.top)
        case Side.LEFT:
            return makePt(b.left, b.top + round(b.height * p.distance))
        case Side.RIGHT:
            return makePt(b.right, b.top + round(b.height * p.distance))
        case _:
            assert False, f'Unknown side {p.side}'


def reducePoints(points: Points) -> Points:
    """

    Args:
        points:

    Returns: Returns a list without repeated points
    """

    result: Points = Points(list(dict.fromkeys(points)))

    return result


def extrudeConnectorPoint(cp: ConnectorPoint, margin: int) -> Point:

    x, y = computePt(p=cp).toTuple()

    match cp.side:
        case Side.TOP:
            return makePt(x, y - margin)
        case Side.RIGHT:
            return makePt(x + margin, y)
        case Side.BOTTOM:
            return makePt(x, y + margin)
        case Side.LEFT:
            return makePt(x - margin, y)
        case _:
            assert False, f'Unknown side {cp.side}'


def getBendDirection(a: Point, b: Point, c: Point) -> BendDirection:

    equalX: bool = a.x == b.x and b.x == c.x
    equalY: bool = a.y == b.y and b.y == c.y

    segment1Horizontal: bool = a.y == b.y
    segment1Vertical:   bool = a.x == b.x
    segment2Horizontal: bool = b.y == c.y
    segment2Vertical:   bool = b.x == c.x

    if equalX is True or equalY is True:
        return BendDirection.NONE

    if not (segment1Vertical or segment1Horizontal) or not (segment2Vertical or segment2Horizontal):
        return BendDirection.NONE

    # Hope this is correct:
    # a if condition else b

    if segment1Horizontal and segment2Vertical:
        # return c.y > b.y ? 's' : 'n';
        return BendDirection.SOUTH if c.y > b.y else BendDirection.NORTH
    if segment1Vertical and segment2Horizontal:
        #  return c.x > b.x ? 'e' : 'w';
        return BendDirection.EAST if c.x > b.x else BendDirection.WEST

    raise ValueError('Nope')


def simplifyPaths(points: Points) -> Points:
    if len(points) <= 2:
        return points

    r: Points = Points([points[0]])
    # TODO:
    # Yeah, yeah, I know the following code is not "Pythonic";  I am doing
    # a blind TypeScript port;  Once I getting converted and appropriately
    # unit test it I can rewrite it;
    #
    for i in range(len(points)):
        cur: Point = points[i]

        if i == len(points) - 1:
            r.append(cur)
            break

        prev: Point = points[i - 1]
        nxt:  Point = points[i + 1]

        bendDirection: BendDirection = getBendDirection(prev, cur, nxt)
        if bendDirection != BendDirection.NONE:
            r.append(cur)

    return r




from typing import Dict
from typing import NewType
from typing import cast

from logging import Logger
from logging import getLogger

from pyorthogonalrouting.Common import Integers
from pyorthogonalrouting.Functions import reducePoints
from pyorthogonalrouting.Point import Points

from pyorthogonalrouting.Rectangle import NO_RECTANGLE
from pyorthogonalrouting.Rectangle import Rectangle
from pyorthogonalrouting.Rectangle import Rectangles

RowNumber    = NewType('RowNumber',    int)
ColumnNumber = NewType('ColumnNumber', int)

RectangleMapRow = NewType('RectangleMapRow', Dict[ColumnNumber, Rectangle])
GridMap         = NewType('GridMap',         Dict[RowNumber, RectangleMapRow])

DEBUG_GRID: bool = False     # TODO:  Make this a runtime flag


class Grid:
    def __init__(self):
        self.logger: Logger = getLogger(__name__)

        self._rows: int = 0
        self._cols: int = 0

        self._gridMap: GridMap = GridMap({})

    @classmethod
    def rulersToGrid(cls, verticals: Integers, horizontals: Integers, bounds: Rectangle) -> 'Grid':
        """
        Creates a grid of rectangles from the specified set of rulers, contained on the specified bounds

        Args:
            verticals:
            horizontals:
            bounds:
        Returns:  The newly created Grid
       """
        grid: Grid = Grid()

        verticals.sort()
        horizontals.sort()

        lastX:  int = bounds.left
        lastY:  int = bounds.top
        column: int = 0
        row:    int = 0

        for y in horizontals:
            for x in verticals:
                rectangle: Rectangle = Rectangle.fromLTRB(left=lastX, top=lastY, right=x, bottom=y)
                grid.set(row=RowNumber(row), column=ColumnNumber(column), rectangle=rectangle)
                column += 1
                lastX = x

            # last cell of the row

            grid.set(row=RowNumber(row), column=ColumnNumber(column), rectangle=Rectangle.fromLTRB(lastX, lastY, bounds.right, y))
            lastX  = bounds.left
            lastY  = y
            column = 0
            row    += 1

        lastX = bounds.left

        # last row of cells
        for x in verticals:
            rectangle = Rectangle.fromLTRB(lastX, lastY, x, bounds.bottom)
            grid.set(row=RowNumber(row), column=ColumnNumber(column), rectangle=rectangle)
            column += 1
            lastX = x

        # Last cell of row
        rectangle = Rectangle.fromLTRB(lastX, lastY, bounds.right, bounds.bottom)
        grid.set(row=RowNumber(row), column=ColumnNumber(column), rectangle=rectangle)

        if DEBUG_GRID is True:
            Grid.debugGridMap(grid=grid)

        return grid

    @classmethod
    def gridToSpots(cls, grid: 'Grid', obstacles: Rectangles) -> Points:

        gridPoints: Points = Points([])

        for row, data in grid._gridMap.items():
            firstRow: bool = row == 0
            lastRow:  bool = row == grid.rows - 1

            for col, r in data.items():
                firstCol: bool = col == 0
                lastCol:  bool = col == grid.columns - 1

                nw: bool = firstCol and firstRow
                ne: bool = firstRow and lastCol
                se: bool = lastRow and lastCol
                sw: bool = lastRow and firstCol
                if nw or ne or se or sw:
                    gridPoints = Points(gridPoints + Points([r.northWest, r.northEast, r.southWest, r.southEast]))
                elif firstRow is True:
                    gridPoints = Points(gridPoints + Points([r.northWest, r.north, r.northEast]))
                elif lastRow is True:
                    gridPoints = Points(gridPoints + Points([r.southEast, r.south, r.southWest]))
                elif firstCol is True:
                    gridPoints = Points(gridPoints + Points([r.northWest, r.west, r.southWest]))
                elif lastCol is True:
                    gridPoints = Points(gridPoints + Points([r.northEast, r.east, r.southEast]))
                else:
                    gridPoints = Points(gridPoints + Points([r.northWest, r.north, r.northEast, r.east, r.southEast, r.south, r.southWest, r.west, r.center]))

        # obstacleCollision = (p: Point) = > obstacles.filter(o= > o.contains(p)).length > 0;
        # filter out points that do not touch shapes
        noCollisions: Points = Rectangle.getNotColliding(points=gridPoints, rectangles=obstacles)
        noDuplicates: Points = reducePoints(noCollisions)

        return noDuplicates

    @property
    def rows(self) -> int:
        return self._rows

    @property
    def columns(self) -> int:
        return self._cols

    @property
    def rectangles(self) -> Rectangles:

        r: Rectangles = Rectangles([])

        for rr in self._gridMap.values():
            rectangleMapRow: RectangleMapRow = cast(RectangleMapRow, rr)
            for rrr in rectangleMapRow.values():
                rectangle: Rectangle = cast(Rectangle, rrr)
                r.append(rectangle)

        return r

    def set(self, row: RowNumber, column: ColumnNumber, rectangle: Rectangle):

        self._rows = max(self.rows,    row + 1)
        self._cols = max(self.columns, column + 1)

        if row in self._gridMap.keys():
            rectangleMapRow: RectangleMapRow = self._gridMap[row]
        else:
            rectangleMapRow = RectangleMapRow({})
            self._gridMap[row] = rectangleMapRow

        rectangleMapRow[column] = rectangle

    def get(self, row: RowNumber, column: ColumnNumber) -> Rectangle:

        if row in self._gridMap.keys():
            rectangleMapRow: RectangleMapRow = self._gridMap[row]
            if column in rectangleMapRow:
                return rectangleMapRow[column]
            else:
                return NO_RECTANGLE
        else:
            return NO_RECTANGLE

    @classmethod
    def debugGridMap(cls, grid: 'Grid'):

        gridMap: GridMap = grid._gridMap

        lc: str = '{'   # left curly
        rc: str = '}'   # right curly

        print(f'GridMap [{grid.rows}] {lc}')
        for x in gridMap.keys():
            rectangleMapRow: RectangleMapRow = gridMap[x]
            print(f'\t{x} => Map[{len(rectangleMapRow)}] {lc}')
            for y in rectangleMapRow.keys():
                rectangle: Rectangle = gridMap[x][y]
                print(f'\t\t{y} => {rectangle}')
            print(f'\t\t{rc},')

        print(f'{rc}')




from typing import List
from typing import NewType

from dataclasses import dataclass

from pyorthogonalrouting.Common import NOT_SET_INT


@dataclass
class Size:
    """
    A size construct
    """
    width:  int = NOT_SET_INT
    height: int = NOT_SET_INT

    def __str__(self):
        return f'{self.width},{self.height}'

    def __repr__(self):
        return self.__str__()


Sizes = NewType('Sizes', List[Size])




from typing import List
from typing import NewType
from typing import cast

from sys import maxsize
from sys import float_info

X_SENTINEL: int = -666
Y_SENTINEL: int = -666

WIDTH_SENTINEL:  int = -777
HEIGHT_SENTINEL: int = -777

NOT_SET_SENTINEL: int = 0xDEADBEEF

NOT_SET_BOOLEAN: bool  = cast(bool, None)
NOT_SET_INT:     int   = cast(int, None)
NOT_SET_FLOAT:   float = cast(float, None)

NO_STRING:       str  = cast(str, None)

INT_MAX:    int   = maxsize
FLOAT_MAX:  float = float_info.max

Integers = NewType('Integers', List[int])


def integerListFactory() -> Integers:
    return Integers([])




from enum import Enum


class BendDirection(Enum):

    NORTH   = 'North'
    EAST    = 'East'
    SOUTH   = 'South'
    WEST    = 'West'
    UNKNOWN = 'Unknown'
    NONE    = 'None'
    NOT_SET = 'Not Set'





from typing import cast
from typing import Dict
from typing import List
from typing import NewType

from dataclasses import dataclass
from dataclasses import field

from pyorthogonalrouting.Common import INT_MAX
from pyorthogonalrouting.Point import NO_POINT
from pyorthogonalrouting.Point import Point

Distance = NewType('Distance', int)

PointNodes    = NewType('PointNodes',    List['PointNode'])
PointNodesMap = NewType('PointNodesMap', Dict['PointNode', Distance])


def pointNodesFactory() -> PointNodes:
    return PointNodes([])


def pointNodesMapFactory() -> PointNodesMap:
    return PointNodesMap({})


@dataclass
class PointNode:
    """
    Represents a node in a graph, whose data is a Point
    """
    distance:      int           = INT_MAX
    shortestPath:  PointNodes    = field(default_factory=pointNodesFactory)
    adjacentNodes: PointNodesMap = field(default_factory=pointNodesMapFactory)
    data:          Point         = NO_POINT

    @classmethod
    def typeScriptCopy(cls, pointNode: 'PointNode') -> 'PointNode':
        """
        Emulate the TypeScript copy mechanism.

        See https://stackoverflow.com/questions/50051149/does-spreading-create-shallow-copy

        Args:
            pointNode:

        Returns:
        """

        doppleGanger: PointNode = PointNode()

        doppleGanger.distance      = pointNode.distance
        doppleGanger.data          = Point(x=pointNode.data.x, y=pointNode.data.y)
        doppleGanger.shortestPath  = pointNode.shortestPath
        doppleGanger.adjacentNodes = pointNode.adjacentNodes

        return doppleGanger

    def __eq__(self, other) -> bool:
        """

        Args:
            other:

        Returns:  True if the defined PointNodes are 'functionally' equal
        """
        ans: bool = False

        if isinstance(other, PointNode) is False:
            pass
        else:
            if self.distance == other.distance and self.data.x == other.data.x and self.data.y == other.data.y:
                ans = True

        return ans

    def __hash__(self) -> int:
        return hash((self.distance, self.data.x, self.data.y))

    def __str__(self) -> str:
        return f'distance={self.distance} Point={self.data}'

    def __repr__(self) -> str:
        return self.__str__()


NO_POINT_NODE: PointNode = cast(PointNode, None)




from typing import cast
from typing import List
from typing import NewType

from dataclasses import dataclass

from pyorthogonalrouting.Common import NOT_SET_FLOAT
from pyorthogonalrouting.Rect import NO_RECT
from pyorthogonalrouting.Rect import Rect
from pyorthogonalrouting.enumerations.Side import Side


@dataclass
class ConnectorPoint:
    """
    Represents a connection point on a routing request
    """
    shape:    Rect  = NO_RECT
    side:     Side  = Side.NOT_SET
    distance: float = NOT_SET_FLOAT


ConnectorPoints = NewType('ConnectorPoints', List[ConnectorPoint])

NO_CONNECTOR_POINT: ConnectorPoint = cast(ConnectorPoint, None)




from typing import cast
from typing import List
from typing import NewType

from logging import Logger
from logging import getLogger

from pyorthogonalrouting.Point import Point
from pyorthogonalrouting.Point import Points
from pyorthogonalrouting.Rect import Rect
from pyorthogonalrouting.Size import Size


class Rectangle:
    """
    Abstracts a Rectangle and adds geometric utilities
    """
    def __init__(self, left: int, top: int, width: int, height: int):

        self.logger: Logger = getLogger(__name__)

        self._left:   int = left
        self._top:    int  = top
        self._width:  int = width
        self._height: int = height

    @classmethod
    def empty(cls) -> 'Rectangle':
        return Rectangle(0, 0, 0, 0)

    @classmethod
    def fromRect(cls, r: Rect) -> 'Rectangle':
        return Rectangle(left=r.left, top=r.top, width=r.width, height=r.height)

    @classmethod
    def fromLTRB(cls, left: int, top: int, right: int, bottom: int) -> 'Rectangle':
        return Rectangle(left=left, top=top, width=right-left, height=bottom-top)

    @classmethod
    def getNotColliding(cls, points: Points, rectangles: 'Rectangles') -> Points:

        notColliding = [pt for pt in points if Rectangle.obstacleCollision(pt, rectangles) is False]

        return Points(notColliding)

    @classmethod
    def obstacleCollision(cls, point: Point, rectangles: 'Rectangles') -> bool:
        """
        Determine if the point is in any of the rectangles

        Args:
            point:
            rectangles:

        Returns:  'True' if the point in one of the rectangles, otherwise 'False'
        """
        ans: bool = False

        for r in rectangles:
            rectangle: Rectangle = cast(Rectangle, r)
            if rectangle.contains(p=point):
                ans = True
                break

        return ans

    @property
    def left(self) -> int:
        return self._left

    @property
    def top(self) -> int:
        return self._top

    @property
    def width(self) -> int:
        return self._width

    @property
    def height(self) -> int:
        return self._height

    @property
    def right(self) -> int:
        return self._left + self._width

    @property
    def bottom(self) -> int:
        return self._top + self._height

    @property
    def center(self) -> Point:
        return Point(x=self.left + self.width // 2, y=self.top + self.height // 2)

    @property
    def location(self) -> Point:
        return Point(x=self.left, y=self.top)

    @property
    def northEast(self) -> Point:
        return Point(x=self.right, y=self.top)

    @property
    def southEast(self) -> Point:
        return Point(x=self.right, y=self.bottom)

    @property
    def southWest(self) -> Point:
        return Point(x=self.left, y=self.bottom)

    @property
    def northWest(self) -> Point:
        return Point(x=self.left, y=self.top)

    @property
    def east(self) -> Point:
        return Point(x=self.right, y=self.center.y)

    @property
    def north(self) -> Point:
        return Point(x=self.center.x, y=self.top)

    @property
    def south(self) -> Point:
        return Point(x=self.center.x, y=self.bottom)

    @property
    def west(self) -> Point:
        return Point(x=self.left, y=self.center.y)

    @property
    def size(self) -> Size:
        return Size(width=self.width, height=self.height)

    # noinspection PyChainedComparisons
    def contains(self, p: Point) -> bool:
        return p.x >= self._left and p.x <= self.right and p.y >= self._top and p.y <= self.bottom

    def inflate(self, horizontal: int, vertical: int) -> 'Rectangle':

        return Rectangle.fromLTRB(left=self._left - horizontal,
                                  top=self._top - vertical,
                                  right=self.right + horizontal,
                                  bottom=self.bottom + vertical)

    def intersects(self, rectangle: 'Rectangle') -> bool:

        thisX: int = self._left
        thisY: int = self._top
        thisW: int = self._width
        thisH: int = self._height
        rectX: int = rectangle.left
        rectY: int = rectangle.top
        rectW: int = rectangle.width
        rectH: int = rectangle.height

        return (rectX < thisX + thisW) and (thisX < (rectX + rectW)) and (rectY < thisY + thisH) and (thisY < rectY + rectH)

    def union(self, r: 'Rectangle') -> 'Rectangle':

        x = [self.left, self.right,  r.left, r.right]
        y = [self.top,  self.bottom, r.top,  r.bottom]

        return Rectangle.fromLTRB(left=min(x), top=min(y), right=max(x), bottom=max(y))

    def __eq__(self, other) -> bool:
        """

        Args:
            other:

        Returns:  True if the defined rectangles are 'functionally' equal
        """
        ans: bool = False

        if isinstance(other, Rectangle) is False:
            pass
        else:
            if self.left == other.left and self.top == other.top and self.width == other.width and self.height == other.height:
                ans = True

        return ans

    def __str__(self) -> str:
        return f'left: {self.left}, top: {self.top}, width: {self.width}, height: {self.height}'

    def __repr__(self) -> str:
        return self.__str__()


Rectangles = NewType('Rectangles', List[Rectangle])

NO_RECTANGLE = cast(Rectangle, None)


def rectanglesFactory() -> Rectangles:
    return Rectangles([])




from typing import Dict
from typing import NewType
from typing import Set
from typing import Tuple
from typing import cast

from logging import Logger
from logging import getLogger

from dataclasses import dataclass
from dataclasses import field
from pyorthogonalrouting.Common import INT_MAX
from pyorthogonalrouting.Common import Integers
from pyorthogonalrouting.Line import Line
from pyorthogonalrouting.Line import Lines
from pyorthogonalrouting.Functions import distance
from pyorthogonalrouting.Line import linesFactory

from pyorthogonalrouting.Point import Point
from pyorthogonalrouting.Point import Points

from pyorthogonalrouting.PointNode import NO_POINT_NODE
from pyorthogonalrouting.PointNode import PointNode
from pyorthogonalrouting.PointNode import PointNodes

from pyorthogonalrouting.enumerations.Direction import Direction

XStr = NewType('XStr', str)
YStr = NewType('YStr', str)

YToPointNodeDict = NewType('YToPointNodeDict', Dict[YStr, PointNode])
XToYDict         = NewType('XToYDict',         Dict[XStr, YToPointNodeDict])

PointNodeSet     = Set[PointNode]
PointNodeDict    = NewType('PointNodeDict', Dict[Point, PointNode])


@dataclass
class GraphAndConnections:
    graph:       'PointGraph' = cast('PointGraph', None)
    connections: Lines        = field(default_factory=linesFactory)


class PointNotFoundException(Exception):
    pass


class PointNodeNotFoundException(Exception):
    pass


class PointGraph:
    """
    Represents a Graph of Point nodes
    """
    def __init__(self):

        self.logger: Logger = getLogger(__name__)

        self._index: XToYDict = XToYDict({})

    @classmethod
    def createGraph(cls, spots: Points) -> GraphAndConnections:
        """
        Creates a graph connecting the specified points orthogonally

        Args:
            spots:

        Returns:
        """
        hotXs:       Integers   = Integers([])
        hotYs:       Integers   = Integers([])
        graph:       PointGraph = PointGraph()
        connections: Lines      = linesFactory()

        for point in spots:
            p: Point = cast(Point, point)
            x, y = p.toTuple()

            if x not in hotXs:
                hotXs.append(x)
            if y not in hotYs:
                hotYs.append(y)
            graph.add(p=p)

        # graph._debugIndex()
        hotXs.sort()
        hotYs.sort()

        def inHotIndex(pt: Point) -> bool:
            return graph.has(pt)

        # TODO:
        # Yeah, yeah, I know the following code is not "Pythonic";  I am doing
        # a blind TypeScript port;  Once I getting converted and appropriately
        # unit test it I can rewrite it;
        #
        for i in range(len(hotYs)):
            for j in range(len(hotXs)):
                b: Point = Point(x=hotXs[j], y=hotYs[i])

                if inHotIndex(b) is False:
                    continue
                if j > 0:
                    a: Point = Point(x=hotXs[j - 1], y=hotYs[i])
                    if inHotIndex(a) is True:
                        graph.connect(a, b)
                        graph.connect(b, a)
                        connections.append(Line(a=a, b=b))
                if i > 0:
                    a = Point(x=hotXs[j], y=hotYs[i - 1])
                    if inHotIndex(a) is True:
                        graph.connect(a, b)
                        graph.connect(b, a)
                        connections.append(Line(a=a, b=b))

        # graph._debugIndex()
        # graph.logger.info(f'{connections}')

        return GraphAndConnections(graph=graph, connections=connections)

    @classmethod
    def shortestPath(cls, graph: 'PointGraph', origin: Point, destination: Point) -> Points:
        """
        TODO: Not unit tested, yet.

        Args:
            graph:
            origin:
            destination:

        Returns:
        """

        originNode:      PointNode = graph.get(origin)
        destinationNode: PointNode = graph.get(destination)

        if originNode is None:
            raise PointNodeNotFoundException(f'Origin Node {origin} not found')
        if destinationNode is None:
            raise PointNodeNotFoundException(f'Destination Node {destination} not found')

        graph.calculateShortestPathFromSource(graph=graph, source=originNode)

        pointNodes: PointNodes = destinationNode.shortestPath
        points:     Points = Points([pointNode.data for pointNode in pointNodes])

        return points

    def add(self, p: Point):

        xs, ys = self._pointToString(p=p)

        if xs not in self._index.keys():
            self._index[xs] = YToPointNodeDict({})

        yToPointNodeDict: YToPointNodeDict = self._index[xs]
        if ys not in yToPointNodeDict.keys():
            yToPointNodeDict[ys] = PointNode(data=p)

        self.logger.debug(f'Add of {p=} complete {self._index=}')

    def has(self, p: Point) -> bool:

        xs, ys = self._pointToString(p=p)

        return xs in self._index and ys in self._index[xs]

    def connect(self, a: Point, b: Point):
        """

        Args:
            a:
            b:
        """
        nodeA: PointNode = self.get(p=a)
        nodeB: PointNode = self.get(p=b)
        if nodeA == NO_POINT_NODE:
            raise PointNotFoundException(f'No Point a: {a=}')
        if nodeB == NO_POINT_NODE:
            raise PointNotFoundException(f'No Point b: {b=}')

        nodeA.adjacentNodes[nodeB] = distance(a=a, b=b)

    def get(self, p: Point) -> PointNode:
        """
        Retrieve the PointNode associated with the Point

        Args:
            p:

        Returns:  May return None
        """
        xs, ys = self._pointToString(p=p)
        if self.has(p) is True:
            return self._index[xs][ys]

        return NO_POINT_NODE

    def calculateShortestPathFromSource(self, graph: 'PointGraph', source: PointNode) -> 'PointGraph':

        source.distance = 0
        settledNodes:   PointNodeSet  = set()
        #
        # TODO:
        # Yeah, had to use a dictionary here to track unsettled nodes;  Seems like a bug in
        # how the set object was removing an incorrect node, which resulted in an error
        # Need to look at this sometime in the future
        #
        unSettledNodes: PointNodeDict = PointNodeDict({})

        unSettledNodes[source.data] = source

        while len(unSettledNodes) != 0:
            currentNode: PointNode = self._getLowestDistanceNode(unSettledNodes)
            # self.logger.info(f'{currentNode=}')
            # unSettledNodes.remove(currentNode)
            del unSettledNodes[currentNode.data]
            # unSettledNodes.discard(currentNode)

            for adjacentNode, edgeWeight in currentNode.adjacentNodes.items():
                if adjacentNode not in settledNodes:
                    self._calculateMinimumDistance(evaluationNode=adjacentNode, edgeWeight=edgeWeight, sourceNode=currentNode)
                    unSettledNodes[adjacentNode.data] = adjacentNode
            settledNodes.add(currentNode)

        return graph

    def _inferPathDirection(self, node: PointNode) -> Direction:
        if len(node.shortestPath) == 0:
            return Direction.UNKNOWN

        return self._directionOfNodes(a=node.shortestPath[len(node.shortestPath) - 1], b=node)

    def _directionOfNodes(self, a: PointNode, b: PointNode) -> Direction:
        return self._directionOf(a=a.data, b=b.data)

    def _directionOf(self, a: Point, b: Point) -> Direction:
        """
        In the original JS version this method return None
        Args:
            a:
            b:

        Returns: Vertical or Horizontal;  May return unknown

        """
        if a.x == b.x:
            return Direction.HORIZONTAL
        elif a.y == b.y:
            return Direction.VERTICAL
        else:
            return Direction.UNKNOWN

    def _getLowestDistanceNode(self, unSettledNodes: PointNodeDict) -> PointNode:

        lowestDistanceNode: PointNode = cast(PointNode, None)
        lowestDistance:     int       = INT_MAX
        for n in unSettledNodes.values():
            node:         PointNode = cast(PointNode, n)
            nodeDistance: int       = node.distance
            if nodeDistance < lowestDistance:
                lowestDistance     = nodeDistance
                lowestDistanceNode = node

        self.logger.debug(f'{lowestDistanceNode=}')
        return lowestDistanceNode

    def _calculateMinimumDistance(self, evaluationNode: PointNode, edgeWeight: int, sourceNode: PointNode):

        sourceDistance:    int       = sourceNode.distance
        comingDirection:   Direction = self._inferPathDirection(sourceNode)
        goingDirection:    Direction = self._directionOfNodes(sourceNode, evaluationNode)
        #
        # Rewrite the following in a simpler to understand mechanism
        # const changingDirection = comingDirection && goingDirection && comingDirection != goingDirection;
        # extraWeigh = changingDirection ? Math.pow(edgeWeigh + 1, 2) : 0;
        if comingDirection == Direction.UNKNOWN or goingDirection == Direction.UNKNOWN:
            changingDirection: bool = False
        else:
            if comingDirection == goingDirection:
                changingDirection = False
            else:
                changingDirection = True
        if changingDirection is True:
            extraWeight: int = pow(edgeWeight + 1, 2)
        else:
            extraWeight = 0

        if sourceDistance + edgeWeight + extraWeight < evaluationNode.distance:

            evaluationNode.distance  = sourceDistance + edgeWeight + extraWeight
            # shortestPath: PointNode[] = [...sourceNode.shortestPath]
            # shortestPath: PointNodes = deepcopy(sourceNode.shortestPath)
            # shortestPath: PointNodes = copy(sourceNode.shortestPath)
            shortestPath: PointNodes = self._typeScriptSpread(sourceNode.shortestPath)

            sourceDoppleGanger: PointNode = PointNode.typeScriptCopy(sourceNode)
            # shortestPath.append(sourceNode)
            shortestPath.append(sourceDoppleGanger)
            evaluationNode.shortestPath = shortestPath

    def _pointToString(self, p: Point) -> Tuple[XStr, YStr]:

        xs: XStr = XStr(str(p.x))
        ys: YStr = YStr(str(p.y))

        return xs, ys

    def _debugIndex(self):

        lc: str = '{'   # left curly
        rc: str = '}'   # right curly

        index: XToYDict = self._index

        prettyIndex: str = f'index: {lc}'
        for xKey in index.keys():
            ytoPointNode: YToPointNodeDict = index[xKey]
            prettyIndex = f'{prettyIndex}\n\t{xKey}: {lc}'

            for yKey in ytoPointNode.keys():
                pointNode: PointNode = ytoPointNode[yKey]
                prettyIndex = f'{prettyIndex}\n\t\t{yKey:>4}: {pointNode}'

            prettyIndex = f'{prettyIndex}\n\t{rc},'

        prettyIndex = f'{prettyIndex}\n{rc}'
        self.logger.info(f'{prettyIndex}')

    def _typeScriptSpread(self, pointNodes: PointNodes) -> PointNodes:
        """
        See:
            https://stackoverflow.com/questions/50051149/does-spreading-create-shallow-copy

        Args:
            pointNodes:

        Returns:
        """

        doppleGanger: PointNodes = PointNodes([])

        for pt in pointNodes:
            duplicate: PointNode = PointNode.typeScriptCopy(pt)
            doppleGanger.append(duplicate)

        return doppleGanger




from dataclasses import dataclass

from pyorthogonalrouting.Common import NOT_SET_INT


@dataclass
class LeftTopRightBottom:

    left:   int = NOT_SET_INT
    top:    int = NOT_SET_INT
    right:  int = NOT_SET_INT
    bottom: int = NOT_SET_INT
